#include <stdio.h>
#include <stdlib.h>
#include <sys/wait.h>
#include <fcntl.h>
#include <string.h>
#include <unistd.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>

#include "messages.pb-c.h"
#define DIM 2048

int ricevi (int sd, char* buf, int n){
   int i, j;
   i = recv (sd, buf, n, 0);
   if ( i != n && i != 0) {
      if (i == -1) { 
         fprintf(stderr,"errore in lettura\n");
          exit(1); }
      while (i < n) { 
         j = recv (sd, &buf[i], n-i, 0);
         if (j == -1) { 
            fprintf(stderr,"errore in lettura\n"); 
            exit(1); }
         i += j; 
         if (j == 0) break; 
      }
   } /* si assume che tutti i byte arrivino… se si verifica il fine file si esce */
   return i;
}

int main(int argc, char** argv){
	int err;
	struct addrinfo hints;
	struct addrinfo *res,*ptr;
	char *host_remoto;
	char *servizio_remoto;
	int sd, nread;
	int i=1;
	
	// Tipi di dato defniti tramite protobuf
	Request req_msg = REQUEST__INIT;
	void *req_buf;      // Buffer to store serialized data
	unsigned len;      // Length of serialized data
	char dim[1];

	Response *res_msg;
	uint8_t res_buf[DIM];


	/* Controllo argomenti */
	if (argc < 3){
		printf("Uso: ./client <server> <porta> <username> <password> <valore>\n");
		exit(1);
	}

	/* Costruzione dell'indirizzo */
	memset(&hints, 0, sizeof(hints));
	hints.ai_family   = AF_INET;
	hints.ai_socktype = SOCK_STREAM;

	/* Risoluzione dell'host */
	host_remoto = argv[1];
	servizio_remoto = argv[2];

	if ((err = getaddrinfo(host_remoto, servizio_remoto, &hints, &res)) != 0) {
		fprintf(stderr, "Errore risoluzione nome: %s\n", gai_strerror(err));
		exit(2);
	}

	for(ptr = res; ptr != NULL; ptr = ptr->ai_next){
		/*se socket fallisce salto direttamente alla prossima iterazione*/
		if ((sd = socket(ptr->ai_family, ptr->ai_socktype, ptr->ai_protocol)) < 0){
			fprintf(stderr,"creazione socket fallita\n");
			continue;
		}
		/*se connect funziona esco dal ciclo*/
		if (connect(sd, ptr->ai_addr, ptr->ai_addrlen) == 0){
			printf("connect riuscita al tentativo %d\n",i);
			break;
		}
		i++;
		close(sd);
	}

	/* Verifica sul risultato restituito da getaddrinfo */
	if (ptr == NULL) {
		fprintf(stderr, "Errore risoluzione nome: nessun indirizzo corrispondente trovato\n");
		exit(3);
	}
	
	// Liberiamo la memoria allocata da getaddrinfo()
	freeaddrinfo(res);
	
	// Preparo il messaggio
	req_msg.username = argv[3];
	req_msg.password = argv[4];
	req_msg.valore = atof(argv[5]);
	
	// Serializzo il messaggio
	len = request__get_packed_size(&req_msg);
	printf("len: %d\n", len);
	req_buf = malloc(len);
	request__pack(&req_msg, req_buf);
	
	/* Mando un singolo byte con la dimensione del messaggio 
		(per ipotesi dimensione del messaggio inferiore a 2^8-1) 
		In generale, con interi a più byte, problemi di 
		1) dimensione in byte (16 bit, 32 bit, 64 bit), 
		2) ordine dei byte (endianness), 
		3) rappresentazione (ad esempio, complemento a 1 o a 2) */
	dim[0] = len;
	if (write(sd, dim, 1) < 0) { 
			perror("write dim[0]");
			exit(EXIT_FAILURE);
	}
	
	// Mando i byte contenenti il messaggio serializzato
	if (write(sd, req_buf, len) < 0) {
		perror("write request");
		exit(5);
	}
	
	// Rilascio della memoria
	free(req_buf);
	
	/* Leggo un singolo byte con la dimensione del messaggio */
	if ((nread = read(sd, dim, 1)) < 0) {
		perror("read");
		exit(EXIT_FAILURE);
	}

	/* Leggo esattamente buff[0] byte. */
	if( ricevi(sd, res_buf, dim[0]) < 0 ){
		perror("read");
		exit(EXIT_FAILURE);
	}
	
	// Deserializzazione
	res_msg = response__unpack(NULL, dim[0], res_buf);
	if (res_msg==NULL){
		perror("errore deserializzazione\n");
		exit(1);
	}
	
	printf("res_msg->intera: %d\n", res_msg->intera);
	printf("res_msg->frazionaria: %d\n", res_msg->frazionaria);
	
	// Rilascio memoria
	response__free_unpacked(res_msg, NULL);
			
	close(sd);
	return 0;
}
