#include <stdio.h>
#include <stdlib.h>
#include <sys/wait.h>
#include <fcntl.h>
#include <string.h>
#include <unistd.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>

#include "messages.pb-c.h"
#define DIM 2048

int main(int argc, char** argv){
	int err;
	struct addrinfo hints;
	struct addrinfo *res,*ptr;
	char *host_remoto;
	char *servizio_remoto;
	int sd, nread;
	int i=1;
	
	// Tipi di dato defniti tramite protobuf
	Request req_msg = REQUEST__INIT;
	void *req_buf;      // Buffer to store serialized data
	unsigned len;      // Length of serialized data

	Response *res_msg;
	uint8_t res_buf[DIM];


	/* Controllo argomenti */
	if (argc < 3){
		printf("Uso: ./client <server> <porta> <username> <password> <valore>\n");
		exit(1);
	}

	/* Costruzione dell'indirizzo */
	memset(&hints, 0, sizeof(hints));
	hints.ai_family   = AF_INET;
	hints.ai_socktype = SOCK_STREAM;

	/* Risoluzione dell'host */
	host_remoto = argv[1];
	servizio_remoto = argv[2];

	if ((err = getaddrinfo(host_remoto, servizio_remoto, &hints, &res)) != 0) {
		fprintf(stderr, "Errore risoluzione nome: %s\n", gai_strerror(err));
		exit(2);
	}

	for(ptr = res; ptr != NULL; ptr = ptr->ai_next){
		/*se socket fallisce salto direttamente alla prossima iterazione*/
		if ((sd = socket(ptr->ai_family, ptr->ai_socktype, ptr->ai_protocol)) < 0){
			fprintf(stderr,"creazione socket fallita\n");
			continue;
		}
		/*se connect funziona esco dal ciclo*/
		if (connect(sd, ptr->ai_addr, ptr->ai_addrlen) == 0){
			printf("connect riuscita al tentativo %d\n",i);
			break;
		}
		i++;
		close(sd);
	}

	/* Verifica sul risultato restituito da getaddrinfo */
	if (ptr == NULL) {
		fprintf(stderr, "Errore risoluzione nome: nessun indirizzo corrispondente trovato\n");
		exit(3);
	}
	
	// Liberiamo la memoria allocata da getaddrinfo()
	freeaddrinfo(res);
	
	// Preparo il messaggio
	req_msg.username = argv[3];
	req_msg.password = argv[4];
	req_msg.valore = atof(argv[5]);
	
	// Serializzo il messaggio
	len = request__get_packed_size(&req_msg);
	printf("len: %d\n", len);
	req_buf = malloc(len);
	request__pack(&req_msg, req_buf);
	
	// Mando i byte contenenti il messaggio serializzato
	if (write(sd, req_buf, len) < 0) {
		perror("write request");
		exit(5);
	}
	
	// Rilascio della memoria
	free(req_buf);
	
	// Ricevo i byte con la risposta serializzata
	memset(&res_buf, 0, sizeof(res_buf));
	nread = read(sd, res_buf, sizeof(res_buf));
	printf("nread: %d\n", nread);
	if ( nread < 0 ){
		perror("read response");
		exit(6);
	}
	
	// Deserializzazione
	res_msg = response__unpack(NULL, nread, res_buf);
	if (res_msg==NULL){
		perror("errore deserializzazione\n");
		exit(1);
	}
	
	printf("res_msg->intera: %d\n", res_msg->intera);
	printf("res_msg->frazionaria: %d\n", res_msg->frazionaria);
	
	// Rilascio memoria
	response__free_unpacked(res_msg, NULL);
			
	close(sd);
	return 0;
}
