#include <stdio.h>
#include <errno.h>
#include <stdlib.h>
#include <sys/wait.h>
#include <fcntl.h>
#include <string.h>
#include <unistd.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>
#include <netinet/in.h>

#include <math.h>
#include "messages.pb-c.h"
#define DIM 2048

/* Gestore del segnale SIGCHLD */
void handler(int s)
{
	int status;
	while (waitpid(-1, &status, WNOHANG) > 0);
}

int main(int argc, char** argv)
{
	struct addrinfo hints, *res;
	int err, sd, ns, on, nread, pid;
	
	Request *req_msg;
	uint8_t req_buf[DIM];
	
	Response res_msg = RESPONSE__INIT;
	void *res_buf;     // Buffer to store serialized data
	unsigned len;      // Length of serialized data
	
	float root_square;
	
	/* Controllo argomenti */
	if(argc < 2){
		printf("Uso: ./server <porta> \n");
		exit(1);
	}
	
	memset(&hints, 0, sizeof(hints));
	hints.ai_family   = AF_UNSPEC;
	hints.ai_socktype = SOCK_STREAM;
	hints.ai_flags    = AI_PASSIVE;

	if ((err = getaddrinfo(NULL, argv[1], &hints, &res)) != 0) {
		fprintf(stderr, "Errore setup indirizzo bind: %s\n", gai_strerror(err));
		exit(2);
	}
	
	if ((sd = socket(res->ai_family, res->ai_socktype, res->ai_protocol)) < 0){
		perror("Errore in socket");
		exit(3);
	}
	
	on = 1;
	if(setsockopt(sd, SOL_SOCKET, SO_REUSEADDR, &on, sizeof(on))<0){
		perror("setsockopt");
		exit(4);
	}
	
	if (bind(sd, res->ai_addr, res->ai_addrlen) < 0) {
		perror("Errore in bind");
		exit(5);
	}
	
	freeaddrinfo(res);
	
	if(listen(sd, SOMAXCONN)<0){
		perror("listen");
		exit(6);
	}
	
	signal(SIGCHLD, handler);
	
	/* Attendo i client... */
	for(;;){
		printf("Server in ascolto...\n");

		if((ns = accept(sd, NULL, NULL)) < 0) {
			/* Ignoro gli errori di tipo interruzione da segnale. */
			if (errno == EINTR)
				continue;
			/* Gestisco tutte le altre tipologie di errore. */
			perror("accept");
			exit(7);
		}

		/* Generazione di un figlio */
		if((pid = fork())<0){
			perror("fork");
			exit(8);
		}
		else if(pid == 0){
			
			/* figlio */
			
			/* Chiudo la socket passiva */
			close(sd);
			
			/* lettura richiesta */
			nread = read(ns, req_buf, sizeof(req_buf));
			printf("nread: %d\n", nread);
			if(nread < 0 ){
				perror("read request");
				exit(9);
			}
			
			// Deserializzazione
			req_msg = request__unpack(NULL, nread, req_buf);
			if (req_msg==NULL){ 
				perror("errore deserializzazione\n");
				exit(1);
			}
			
			printf("req_msg->username: %s\n", req_msg->username);
			printf("req_msg->password: %s\n", req_msg->password);
			printf("req_msg->valore: %f\n", req_msg->valore);
			
			// sqrt per la radice quadrata
			root_square = sqrt(req_msg->valore);
			res_msg.intera = (int)root_square;
			res_msg.frazionaria = (int)((root_square - res_msg.intera) * 100);
			printf("root_square: %f\n", root_square);
			printf("res_msg.intera: %d\n", res_msg.intera);
			printf("res_msg.frazionaria: %d\n", res_msg.frazionaria);
			
			// Rilascio memoria
			request__free_unpacked(req_msg, NULL);
			
			// Serializzo il messaggio
			len = response__get_packed_size(&res_msg);
			printf("len: %d\n", len);
			res_buf = malloc(len);
			response__pack(&res_msg, res_buf);
	
			// Invio il buffer con la risposta
			if(write(ns, res_buf, len)<0){
				perror("write res_buf");
				exit(10);
			}
			
			// Rilascio memoria
			free(res_buf);
			
			close(ns);
			exit(14);
		}
		// padre
		close(ns);
	}
	return 0;
}
