#include <stdio.h>
#include <stdlib.h>
#include <sys/wait.h>
#include <fcntl.h>
#include <string.h>
#include <unistd.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>

#include "messages-struct.h"

int main(int argc, char** argv){
	int err;
	struct addrinfo hints;
	struct addrinfo *res,*ptr;
	char *host_remoto;
	char *servizio_remoto;
	int sd;
	int i=1;
	
	struct Request req_msg;
	struct Response res_msg;


	/* Controllo argomenti */
	if (argc < 3){
		printf("Uso: ./client <server> <porta> <username> <password> <valore>\n");
		exit(1);
	}

	/* Costruzione dell'indirizzo */
	memset(&hints, 0, sizeof(hints));
	hints.ai_family   = AF_INET;
	hints.ai_socktype = SOCK_STREAM;

	/* Risoluzione dell'host */
	host_remoto = argv[1];
	servizio_remoto = argv[2];

	if ((err = getaddrinfo(host_remoto, servizio_remoto, &hints, &res)) != 0) {
		fprintf(stderr, "Errore risoluzione nome: %s\n", gai_strerror(err));
		exit(2);
	}

	for(ptr = res; ptr != NULL; ptr = ptr->ai_next){
		/*se socket fallisce salto direttamente alla prossima iterazione*/
		if ((sd = socket(ptr->ai_family, ptr->ai_socktype, ptr->ai_protocol)) < 0){
			fprintf(stderr,"creazione socket fallita\n");
			continue;
		}
		/*se connect funziona esco dal ciclo*/
		if (connect(sd, ptr->ai_addr, ptr->ai_addrlen) == 0){
			printf("connect riuscita al tentativo %d\n",i);
			break;
		}
		i++;
		close(sd);
	}

	/* Verifica sul risultato restituito da getaddrinfo */
	if (ptr == NULL) {
		fprintf(stderr, "Errore risoluzione nome: nessun indirizzo corrispondente trovato\n");
		exit(3);
	}
	
	// Liberiamo la memoria allocata da getaddrinfo()
	freeaddrinfo(res);
	
	// preparo la struct
	memcpy(req_msg.username, argv[3], strlen(argv[3])+1 );
	memcpy(req_msg.password, argv[4], strlen(argv[4])+1 );
	req_msg.valore = atof(argv[5]);
	
	// Mando la struct contenente la richiesta
	if (write(sd, &req_msg, sizeof(Request)) < 0) {
		perror("write request");
		exit(5);
	}
	
	// Ricevo la struct contenente la risposta
	if (read(sd, &res_msg, sizeof(Response)) < 0){
		perror("read response");
		exit(6);
	}
	printf("res_msg.intera: %d\n", res_msg.intera);
	printf("res_msg.frazionaria: %d\n", res_msg.frazionaria);
	
	close(sd);
	return 0;
}
